# id3fs - a FUSE-based filesystem for browsing audio metadata
# Copyright (C) 2010  Ian Beckwith <ianb@erislabs.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

package ID3FS::Path::Node;

use strict;
use warnings;

require Exporter;
use vars qw(@ISA @EXPORT $TYPE_BOOL $TYPE_TAG $TYPE_ARTIST $TYPE_ALBUM $TYPE_FILE);
@ISA=qw(Exporter);
@EXPORT=qw($TYPE_BOOL $TYPE_TAG $TYPE_ARTIST $TYPE_ALBUM $TYPE_FILE);
($TYPE_BOOL, $TYPE_TAG, $TYPE_ARTIST, $TYPE_ALBUM, $TYPE_FILE)=(1..5);

sub new
{
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self={};
    bless($self,$class);

    my $db=shift;
    $self->{type}=shift;
    $self->{name}=shift;
    $self->{parents_id}=shift;
    if($self->{type} != $TYPE_BOOL)
    {
	my $table='';
	if   ($self->{type} == $TYPE_TAG)    { $table="tags";    }
	elsif($self->{type} == $TYPE_ARTIST) { $table="artists"; }
	elsif($self->{type} == $TYPE_ALBUM)  { $table="albums";  }
	elsif($self->{type} == $TYPE_FILE)   { $table="files";   }
	$self->{id}=$db->lookup_id($table, $self->{name}, $self->{parents_id});
	return undef unless(defined($self->{id}));
    }
    return $self;
}

sub set
{
    my($self, $name, $val)=@_;
    if(defined($val))
    {
	$self->{$name}=$val;
    }
    return $self->{$name};
}

sub left       { return shift->set("left",       shift); }
sub right      { return shift->set("right",      shift); }
sub name       { return shift->set("name",       shift); }
sub type       { return shift->set("type",       shift); }
sub id         { return shift->set("id",         shift); }
sub parents_id { return shift->set("parents_id", shift); }

sub to_sql
{
    my($self, $hasvals, $not, @joins)=@_;
    $not=0 unless(defined($not));
    my @outjoins=();
    unless(@joins)
    {
	@outjoins = @joins = ("INNER");
    }
    my $str='';

    if($self->type() != $TYPE_BOOL)
    {
	print $self->{id}, " = ", $self->{name}, "\n";
	$str .= "t" . scalar(@joins) . ".id='" . $self->{id} . "'";
	if($not && !$hasvals)
	{
	    $str = "(" . $str . " AND fxt" . scalar(@joins) . ".files_id IS NULL)";
	}
	return ($str, @outjoins);
    }

    my $left=$self->left();
    my $right=$self->right();
    return ("", @outjoins) unless($left || $right);

    my ($leftstr, @leftjoins) = $left->to_sql($hasvals, $not, @joins) if($left);
    push(@joins, @leftjoins);
    push(@outjoins, @leftjoins);

    my $op=$self->name();
    if(defined($op))
    {
	# if we are ANDing add an inner join
	if($op eq "AND")
	{
	    # if right child is a NOT, we don't need extra join/brackets
	    # NOT will do the same and we will end up with an extra one
#	    unless($right && $right->name() && $right->name() eq "NOT")
#	    {
		push(@joins, "INNER");
		push(@outjoins, "INNER");
#	    }
	}
	elsif($op eq "NOT")
	{
	    $not=1;
	    # return an extra join, but don't pass it down to $right->to_sql
	    # if we are looking for a tag *value*, use INNER join rather than LEFT
	    push(@outjoins, ($hasvals ? "INNER" : "LEFT"));
	}
    }
    my ($rightstr, @rightjoins) = $right->to_sql($hasvals, $not, @joins) if($right);
    push(@outjoins, @rightjoins);
    $str = $leftstr;
    $str .= " $op " if($op && !$not);
    $str .= $rightstr;
    $str=("(" . $str . ")") if($op && $left && $right);
#    print "LEFT (", scalar(@$leftjoins), "): ";
#    print $leftstr if $leftstr;
#    print "\n";
#    print "OP: $op\n" if $op;
#    print "RIGHT (", scalar(@$rightjoins), "): ";
#    print $rightstr if $rightstr;
#    print "\n";
    print "OUT(", scalar(@outjoins), "): $str\n";
    return($str, @outjoins);
}

sub used_tags
{
    my($self)=@_;
    if($self->type() == $TYPE_BOOL)
    {
	my @used=();
	push(@used, $self->left()->used_tags())  if($self->left());
	push(@used, $self->right()->used_tags()) if($self->right());
	return(grep { defined; } @used);
    }
    return $self->id();
}

# does the bottom right-most expression end in a NOT?
sub right_ends_in_not
{
    my($self, $node)=@_;
    return 0 unless($node);
    my $right=$node->right();
    if($right && $right->type() == $TYPE_BOOL)
    {
	return $self->right_ends_in_not($right);
    }
    my $op=$node->name();
    return 0 unless($op);
    return 1 if($op eq "NOT");
    return 0;
}

1;
